@@ -89,7 +89,7 @@ void speedy_cb_write(CopyBuf *bp) {
 	    speedy_cb_free(bp);
 	}
     }
-    else if (n == -1) {
+    else if (n == -1 && errno != EAGAIN) {
 	bp->sz = 0;
     }
 }
